from typing import Type

from youtube_clipper.parsers.model import SubtitleParser
from youtube_clipper.parsers.srt import SRTParser
from youtube_clipper.parsers.vtt import VTTParser


PARSERS_REGISTRY: dict[str, Type[SubtitleParser]] = {
    '.srt': SRTParser,
    '.vtt': VTTParser,
}



import argparse
import tempfile

import colorama
from colorama import Fore

from youtube_clipper.downloader import DownloaderLoggingLevel, SubtitlesDownloader
from youtube_clipper.searcher import DeduplicationMode, SubtitlesSearcher
from youtube_clipper.utils import get_available_formats, get_url_from_filename


def main() -> None:
    parser = argparse.ArgumentParser('youtube-clipper', description='YouTube Clipper - tool for finding youtube clips')
    parser.add_argument(
        '--url',
        required=True,
        help='can be any youtube URL supported by yt-dlp: a video, a channel, a playlist, etc; '
             'in case of multiple videos in the output, search will be performed for each video separately',
    )
    parser.add_argument('--query', required=True, help='query to search for')

    yt_dlp_args = parser.add_argument_group('yt-dlp arguments')
    yt_dlp_args.add_argument('--formats', nargs='*', help='subtitles formats to download')
    yt_dlp_args.add_argument('--language', default='en', help='subtitles language')
    yt_dlp_args.add_argument('--subs-dir', required=False, help='if specified, subs will be saved to a provided dir')
    yt_dlp_args.add_argument(
        '--allow-autogenerated', action='store_true', help='whether to allow downloading autogenerated subs'
    )
    yt_dlp_args.add_argument(
        '--logging-level', default='disabled', type=DownloaderLoggingLevel, help='logging level for yt-dlp'
    )

    searcher_args = parser.add_argument_group('searcher arguments')
    searcher_args.add_argument(
        '--show-scores',
        action='store_true',
        help='if set, will show a search score next to a link',
    )
    searcher_args.add_argument('--search-limit', required=False, type=int, help='limit for search results per video')
    searcher_args.add_argument(
        '--enable-pairwise-group',
        action='store_true',
        help='if set, will merge successive overlapping subtitle pairs into one',
    )
    searcher_args.add_argument(
        '--deduplication-mode',
        default='keep_first',
        type=DeduplicationMode,
        help='deduplication mode for consecutive subtitle matches',
    )

    args = parser.parse_args()
    colorama.init(autoreset=True)

    if args.allow_autogenerated is False and args.enable_pairwise_group is False:
        print(Fore.YELLOW + 'Warning: pairwise grouping is disabled on manual subtitles, search might be inaccurate. '
                            'Consider enabling the pairwise grouping with the --enable-pairwise-group option. '
                            'For more details, check README.md')
    available_formats = get_available_formats()
    formats = args.formats or available_formats
    unsupported_formats = set(formats) - set(available_formats)
    if unsupported_formats:
        print(Fore.YELLOW + 'Warning: some of provided formats aren\'t supported:', *unsupported_formats)

    with tempfile.TemporaryDirectory() as tempdir:
        subs_dir = args.subs_dir or tempdir
        downloader = SubtitlesDownloader(
            language=args.language,
            formats=formats,
            allow_autogenerated=args.allow_autogenerated,
            output_dir=subs_dir,
            logging_level=args.logging_level,
        )

        for filename in downloader.get_subtitles(args.url):
            video_url = get_url_from_filename(filename)
            print(Fore.YELLOW + f'Searching for `{args.query}` in {video_url}')

            searcher = SubtitlesSearcher(
                index_directory=tempdir,
                limit=args.search_limit,
                enable_pairwise_group=args.enable_pairwise_group,
                deduplication_mode=args.deduplication_mode,
            )
            searcher.add_subtitles(filename)
            results = searcher.search(args.query)
            if not results:
                print(Fore.RED + f'`{args.query}` wasn\'t found in {video_url}')
                if args.allow_autogenerated and not args.enable_pairwise_group:
                    print(Fore.YELLOW + 'Pairwise grouping was disabled for this search, consider enabling it '
                                        'with the --enable-pairwise-group option. For more details, check README.md')
            else:
                print(Fore.GREEN + f'Found {len(results)} (approximate) occurrence(s) of `{args.query}` in {video_url}')
                for result in results:
                    score_str = Fore.RESET + f'(score: {result.score})' if args.show_scores else ''
                    print(Fore.GREEN + f'{video_url}&t={int(result.offset)}s {score_str}')


if __name__ == '__main__':
    main()



from enum import Enum
from itertools import pairwise
from pathlib import Path

import attr
from whoosh.fields import NUMERIC, TEXT, Schema
from whoosh.index import Index, create_in
from whoosh.searching import Results
from whoosh.qparser import OrGroup, QueryParser

from youtube_clipper.parsers.model import Subtitle
from youtube_clipper.parsers.registry import PARSERS_REGISTRY


# 1-to-1 correspondence with youtube_clipper.parsers.model:Subtitle
SEARCH_SCHEMA = Schema(id=NUMERIC(stored=True), offset=NUMERIC(stored=True), content=TEXT)


@attr.s
class SearchResult:
    offset: float = attr.ib()
    score: float = attr.ib()


class DeduplicationMode(Enum):
    DISABLE = 'disable'
    KEEP_FIRST = 'keep_first'
    KEEP_LAST = 'keep_last'


@attr.s
class SubtitlesSearcher:
    index_directory: str = attr.ib()
    limit: int | None = attr.ib(default=None)

    enable_pairwise_group: bool = attr.ib(default=False)
    deduplication_mode: DeduplicationMode = attr.ib(default=DeduplicationMode.KEEP_FIRST)

    index: Index = attr.ib(init=False, default=attr.Factory(
        lambda self: create_in(self.index_directory, SEARCH_SCHEMA), takes_self=True,
    ))

    def _get_query_parser(self) -> QueryParser:
        """
        See https://whoosh.readthedocs.io/en/latest/parsing.html#common-customizations
        """
        og = OrGroup.factory(0.9)
        return QueryParser('content', self.index.schema, group=og)

    def _normalize(self, text: str) -> str:
        return text.lower().translate(str.maketrans('', '', '.,!?'))

    def parse_results(self, results: Results) -> list[SearchResult]:
        """
        Parse results from searcher and perform a deduplication depending on a `self.deduplication_mode`:
        * DeduplicationMode.DISABLE - just parse results without a deduplication
        * DeduplicationMode.KEEP_FIRST - keep only the first subtitle in every chain of consecutive matches
        * DeduplicationMode.KEEP_LAST - keep only the last subtitle in every chain of consecutive matches
        """
        if not results or self.deduplication_mode == DeduplicationMode.DISABLE:  # skip deduplication
            return [SearchResult(offset=result['offset'], score=result.score) for result in results]

        ids = sorted(result['id'] for result in results)
        kept_ids: set[int] = {ids[0]}  # to cover a corner case of len(results) == 1
        removed_ids: set[int] = set()
        for id_pair in pairwise(ids):  # type: tuple[int, int]
            if id_pair[0] + 1 == id_pair[1]:  # consecutive subtitles - perform a deduplication
                if self.deduplication_mode == DeduplicationMode.KEEP_FIRST:
                    kept_ids.add(id_pair[0])
                    removed_ids.add(id_pair[1])
                elif self.deduplication_mode == DeduplicationMode.KEEP_LAST:
                    removed_ids.add(id_pair[0])
                    kept_ids.add(id_pair[1])
                else:
                    raise ValueError(f'Unknown deduplication mode {self.deduplication_mode}')
            else:  # non-consecutive subtitles - keep both
                kept_ids.add(id_pair[0])
                kept_ids.add(id_pair[1])

        return [  # preserving the original order
            SearchResult(offset=result['offset'], score=result.score)
            for result in results if result['id'] in kept_ids and result['id'] not in removed_ids
        ]

    def add_subtitles(self, filename: str) -> None:
        """
        Parses a subtitles file and adds subtitles to the search index
        Parser is chosen based on a file extension
        If enable_pairwise_group is True, subtitles will be joined in pairs
        """
        writer = self.index.writer()
        parser = PARSERS_REGISTRY[Path(filename).suffix]
        parsed_subtitles = parser().parse_subtitles(filename)

        if self.enable_pairwise_group:
            for subtitles_pair in pairwise(parsed_subtitles):  # type: tuple[Subtitle, Subtitle]
                joined_subtitle = Subtitle(
                    id=subtitles_pair[0].id,
                    offset=subtitles_pair[0].offset,
                    content=self._normalize(' '.join(sub.content for sub in subtitles_pair)),
                )
                writer.add_document(**attr.asdict(joined_subtitle))
        else:
            for subtitle in parsed_subtitles:
                subtitle.content = self._normalize(subtitle.content)
                writer.add_document(**attr.asdict(subtitle))

        writer.commit()

    def search(self, query_string: str) -> list[SearchResult]:
        """
        Perform a search on previously added subtitles
        """
        query_parser = self._get_query_parser()
        query = query_parser.parse(self._normalize(query_string))

        with self.index.searcher() as searcher:
            results = searcher.search(query, limit=self.limit)
            return self.parse_results(results)

    def clear(self) -> None:
        """
        Clear the current index of all previously added documents
        """
        writer = self.index.writer()
        for docnum in range(self.index.doc_count()):
            writer.delete_document(docnum)
        writer.commit()
        assert self.index.doc_count() == 0



import logging
import xml.etree.ElementTree as ET

import ttconv.imsc.reader as imsc_reader
import ttconv.srt.writer as srt_writer

from youtube_clipper.converters.model import SubtitlesConverter


LOGGER = logging.getLogger(__name__)
logging.getLogger('ttconv').setLevel(logging.ERROR)  # too much noise on the WARNING level


class TTMLToSRTConverter(SubtitlesConverter):
    ext_from = '.ttml'
    ext_to = '.srt'

    def _convert(self, source_filename: str, output_filename: str) -> None:
        """
        See ttconv docs:
        https://github.com/sandflow/ttconv/blob/master/doc/imsc_reader.md
        https://github.com/sandflow/ttconv/blob/master/doc/srt_writer.md
        """

        xml_doc = ET.parse(source_filename)
        LOGGER.info(f'Loading a ttml model for {source_filename}')
        doc = imsc_reader.to_model(xml_doc)
        LOGGER.info(f'Converting {source_filename} to srt')
        with open(output_filename, 'w') as f:
            print(srt_writer.from_model(doc), file=f)
        LOGGER.info(f'Conversion completed!')



import logging
from enum import Enum
from pathlib import Path
from typing import Generator

import attr
from yt_dlp import YoutubeDL
from yt_dlp.extractor import gen_extractor_classes

from youtube_clipper.converters.exc import NoCorrespondingParserException
from youtube_clipper.converters.registry import CONVERTERS_REGISTRY
from youtube_clipper.parsers.registry import PARSERS_REGISTRY


LOGGER = logging.getLogger(__name__)


class DownloaderLoggingLevel(Enum):
    DISABLED = 'disabled'
    ENABLED = 'enabled'
    VERBOSE = 'verbose'


@attr.s
class SubtitlesDownloader:
    language: str = attr.ib()
    formats: list[str] = attr.ib(validator=attr.validators.min_len(1))
    output_dir: str = attr.ib()
    allow_autogenerated: bool = attr.ib(default=True)
    logging_level: DownloaderLoggingLevel = attr.ib(default=DownloaderLoggingLevel.DISABLED)

    _converted_dir: str = attr.ib(init=False)

    def __attrs_post_init__(self) -> None:
        converted_dir = Path(self.output_dir) / 'converted'
        converted_dir.mkdir(parents=True, exist_ok=True)  # ensure the dir exists
        self._converted_dir = str(converted_dir)

    def download_subtitles(self, url: str) -> None:
        """
        Downloads subtitles with provided parameters using yt-dlp
        """
        youtube_dl_params = dict(
            skip_download=True,  # don't download a video itself
            writesubtitles=True,  # download subtitles
            writeautomaticsub=self.allow_autogenerated,  # whether to include automatic subtitles
            subtitlesformat='/'.join(self.formats),
            subtitleslangs=[self.language],
            outtmpl=f'{self.output_dir}/%(id)s.%(ext)s',
            verbose=self.logging_level == DownloaderLoggingLevel.VERBOSE,
            quiet=self.logging_level == DownloaderLoggingLevel.DISABLED,
        )
        LOGGER.info(f'YoutubeDL params: {youtube_dl_params}')

        with YoutubeDL(youtube_dl_params, auto_init=False) as ydl:
            extractors = gen_extractor_classes()
            for extractor in extractors:  # keep only youtube extractors
                if extractor.__name__.startswith('Youtube'):
                    ydl.add_info_extractor(extractor)
            ydl.extract_info(url=url)

    def get_subtitles(self, url: str) -> Generator[str, None, None]:
        """
        Downloads subtitles and checks them against available parsers,
        performing conversion if necessary
        """
        self.download_subtitles(url)
        for path in Path(self.output_dir).iterdir():
            if not path.is_file():
                continue

            ext = path.suffix
            if ext in CONVERTERS_REGISTRY:  # convert subtitles and return a new filename
                converter = CONVERTERS_REGISTRY[ext]
                if converter.ext_to not in PARSERS_REGISTRY:
                    raise NoCorrespondingParserException(
                        f'Converter {converter.__name__} doesn\'t have a corresponding parser'
                    )
                LOGGER.info(f'Going to convert {path} with a {converter.__name__}')
                converted_filename = converter(output_dir=self._converted_dir).convert(str(path))
                yield converted_filename
            elif ext in PARSERS_REGISTRY:  # can be parsed right away
                yield str(path)
            else:
                LOGGER.warning(f'Unknown extension {ext} for a file {path}')



from typing import Type

from youtube_clipper.converters.model import SubtitlesConverter
from youtube_clipper.converters.ttml_to_srt import TTMLToSRTConverter


CONVERTERS_REGISTRY: dict[str, Type[SubtitlesConverter]] = {
    '.ttml': TTMLToSRTConverter,
}



from pathlib import Path

from youtube_clipper.converters.registry import CONVERTERS_REGISTRY
from youtube_clipper.parsers.registry import PARSERS_REGISTRY


YOUTUBE_PREFIX = 'https://www.youtube.com/watch'


def get_url_from_filename(filename: str) -> str:
    """
    Path(filename).stem would only remove the last extension, and
    a file from SubtitlesDownloader.download_subtitles would typically
    have multiple, e.g. *.en.ttml
    """
    path = Path(filename)
    video_id = path.name.removesuffix(''.join(path.suffixes))
    return f'{YOUTUBE_PREFIX}?v={video_id}'


def get_available_formats() -> list[str]:
    """Get all available formats from converters and parsers registries"""
    return list(map(lambda ext: ext.removeprefix('.'), CONVERTERS_REGISTRY.keys() | PARSERS_REGISTRY.keys()))



import abc
from pathlib import Path
from typing import ClassVar

import attr

from youtube_clipper.converters.exc import ExtensionError


@attr.s
class SubtitlesConverter(abc.ABC):
    ext_from: ClassVar[str]
    ext_to: ClassVar[str]

    output_dir: str = attr.ib()

    @abc.abstractmethod
    def _convert(self, source_filename: str, output_filename: str) -> None:
        """
        Main implementation, should read subtitles from `source_filename` and
        output the result to `output_filename`
        """
        pass

    def convert(self, source_filename: str) -> str:
        """
        Converts subtitles from `source_filename` with `ext_from` extension
        to a corresponding file with `ext_to` extension
        """
        path = Path(source_filename)
        if path.suffix != self.ext_from:
            raise ExtensionError(f'Expected {self.ext_from} file, got {source_filename}')
        output_filename = f'{self.output_dir}/{path.stem}{self.ext_to}'
        self._convert(source_filename, output_filename)
        return output_filename



import srt

from youtube_clipper.parsers.model import Subtitle, SubtitleParser


class SRTParser(SubtitleParser):
    def parse_subtitles(self, filename: str) -> list[Subtitle]:
        with open(filename) as f:
            subtitles = srt.parse(f.read())
        return [
            Subtitle(id=subtitle.index, offset=subtitle.start.total_seconds(), content=subtitle.content)
            for subtitle in subtitles
        ]



import webvtt

from youtube_clipper.parsers.model import Subtitle, SubtitleParser


class VTTParser(SubtitleParser):
    @staticmethod
    def _parse_timestamp(timestamp: str) -> float:
        h, m, s = timestamp.split(':', maxsplit=3)
        return 3600 * int(h) + 60 * int(m) + float(s)

    def parse_subtitles(self, filename: str) -> list[Subtitle]:
        return [
            Subtitle(id=idx, offset=self._parse_timestamp(caption.start), content=caption.text)
            for idx, caption in enumerate(webvtt.read(filename))
        ]



import abc

import attr


@attr.s
class Subtitle:
    id: int = attr.ib()
    offset: float = attr.ib()
    content: str = attr.ib()


class SubtitleParser(abc.ABC):
    @abc.abstractmethod
    def parse_subtitles(self, filename: str) -> list[Subtitle]:
        pass
